/* 
   jplay2 - JACK audio player for multimedia files

   Copyright (C) 2007 Robin Gareus

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 
   as published by the Free Software Foundation;

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "player.h"
#include "buffer.h"
#include "xmalloc.h"


extern int m_filetype;
extern int m_samplerate;
extern int m_channels;

extern jack_nframes_t m_length;

extern int want_verbose;
extern int want_quiet;
extern float want_speed;

#ifdef HAVE_FFMPEG
/////// FFMPEG  ///////
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>

AVFormatContext* m_formatContext;
AVCodecContext*  m_codecContext;
AVCodec*         m_codec;
AVPacket         m_packet;
int              m_audioStream;

void openfile_ff (char *filename) {

	av_register_all();
	avcodec_init();
	avcodec_register_all();

	AVInputFormat *file_iformat = av_find_input_format( filename );
	
	if ( av_open_input_file( &m_formatContext, filename, file_iformat, 0, NULL ) != 0 ) {
		fprintf(stderr, "This file can not be read by ffmpeg.\n" );
		return;
	}

	if ( av_find_stream_info( m_formatContext ) < 0 ) {
		fprintf(stderr, "av_find_stream_info failed\n" );
		return;
	}
	m_audioStream = -1;
	int i;
	for (i = 0; i < m_formatContext->nb_streams; i++ ) {
		if ( m_formatContext->streams[i]->codec->codec_type == CODEC_TYPE_AUDIO ) {
			m_audioStream = i;
			break;
		}
	}
	if ( m_audioStream == -1 ) {
		fprintf(stderr, "No Audio Stream found in file\n" );
		return;
	}
	m_codecContext = m_formatContext->streams[m_audioStream]->codec;
	m_codec = avcodec_find_decoder( m_codecContext->codec_id );
	if ( m_codec == NULL ) {
		fprintf(stderr, "Codec not supported by ffmpeg\n" );
		return;
	}
	if ( avcodec_open( m_codecContext, m_codec ) < 0 ) {
		fprintf(stderr, "avcodec_open failed\n" );
		return;
	}

	printf("ffmpeg - audio tics: %i/%i [sec]\n",m_formatContext->streams[m_audioStream]->time_base.num,m_formatContext->streams[m_audioStream]->time_base.den);
	m_samplerate = m_codecContext->sample_rate;
	m_channels= m_codecContext->channels ;
	int64_t len = m_formatContext->duration - m_formatContext->start_time;
	m_length = (int64_t)( len * m_samplerate / AV_TIME_BASE );
	m_filetype = SFT_FFMPEG;
}

int16_t          m_tmpBuffer[AVCODEC_MAX_AUDIO_FRAME_SIZE];
int16_t*         m_tmpBufferStart = 0;
unsigned long    m_tmpBufferLen = 0;
int64_t 	 m_pts = 0;
double audio_clock = 0.0;
double ffjb_clock = 0.0;



void sampleseek_ff (jack_nframes_t sample) {
	// sequential only :(
	if (sample == 0) {
		av_seek_frame( m_formatContext, m_audioStream, 0, AVSEEK_FLAG_ANY| AVSEEK_FLAG_BACKWARD);
		m_tmpBufferLen = 0;
		audio_clock = 0.0;
		return;
	}

	m_tmpBufferLen = 0;
	int64_t timestamp = (int64_t) sample * (int64_t) m_formatContext->streams[m_audioStream]->time_base.den / (int64_t) m_formatContext->streams[m_audioStream]->time_base.num / (int64_t)m_samplerate;
	av_seek_frame( m_formatContext, m_audioStream, timestamp, AVSEEK_FLAG_BACKWARD);
	audio_clock = -1.0; // get from first next frame.
}

// TODO: dither samples.
void decode_int16_to_float(void * _in, jack_default_audio_sample_t ** out, int num_channels, int num_samples, int out_offset)
{
  int i, j;
  int16_t * in;
  for(i = 0; i < num_channels; i++)
    {
    if(out[i])
      {
      in = ((int16_t*)_in) + i;
      for(j = out_offset; j < out_offset+num_samples; j++)
        {	
		int16_t sval = (*in);
		if (sval>=0) out[i][j] = (float)  sval / 32768.0;
		else out[i][j] = (float) sval / 32767.0 ;
                //if (1) out[i][j] = 32768 - out[i][j]; // invert / test
		in += num_channels;
        }
      }
    }
}


jack_nframes_t fillBuffer_ff ( jack_default_audio_sample_t ** bufferptrs, jack_nframes_t frames ) {
//	int i;

	unsigned long written = 0;
	int ret = 0;

	while ( ret >= 0 && written < frames ) {
		if ( m_tmpBufferLen > 0 ) {
			int s = min( m_tmpBufferLen / m_codecContext->channels, frames - written );
			decode_int16_to_float( m_tmpBufferStart, bufferptrs, m_codecContext->channels, s , written);
			written += s;
//			printf("DP: %i - %lu/%u\n",s,written,frames);
			s = s * m_codecContext->channels;
			m_tmpBufferStart += s;
			m_tmpBufferLen -= s;
			ret = 0;
		} else {
			ret = av_read_frame( m_formatContext, &m_packet );
			if (ret <0  && want_verbose) { printf("reached end of file.\n"); break; }

			int len = m_packet.size;
			uint8_t *ptr = m_packet.data;

			while ( ptr != NULL && ret >= 0 && m_packet.stream_index == m_audioStream && len > 0 ) {
                                //int data_size= AVCODEC_MAX_AUDIO_FRAME_SIZE - m_tmpBufferLen - (m_tmpBufferStart-m_tmpBuffer);
                                int data_size= AVCODEC_MAX_AUDIO_FRAME_SIZE;
#if LIBAVCODEC_VERSION_MAJOR < 52
				ret = avcodec_decode_audio2( m_codecContext, m_tmpBuffer, &data_size, ptr, len );
#else
				ret = avcodec_decode_audio3( m_codecContext, m_tmpBuffer, &data_size, &m_packet );
#endif
				if ( ret < 0 ) {
					ret = 0;
					break;
				}
				len -= ret;
				ptr += ret;
//				printf("DATASIZE : %i ++ \n",data_size);

				/* if update the audio clock with the pts */
				if (audio_clock < 0.0 && m_packet.pts != AV_NOPTS_VALUE) {
					audio_clock = m_codecContext->sample_rate * av_q2d(m_formatContext->streams[m_audioStream]->time_base)*m_packet.pts;
				}
				//audio_clock += (double)data_size / (2.0 *  m_codecContext->channels * m_codecContext->sample_rate);
				audio_clock += (double)data_size / (2.0 *  m_codecContext->channels ); // FIXME!
//				printf("CLK: %.0f %.2f\n",audio_clock,audio_clock/m_codecContext->sample_rate);


				if ( data_size > 0 ) {
					m_tmpBufferStart = m_tmpBuffer;
					m_tmpBufferLen = data_size / 2; // 16 bit words.  
				}
			}
			av_free_packet( &m_packet );
		}
	}
	ffjb_clock=audio_clock-(m_tmpBufferLen/ m_codecContext->channels);
//	printf("CLK: %.0f\n",ffjb_clock);
	return written;
}


///////////////////////////////////////////////////
// FFMPEG via buffer 
// sequentially decodes audio 
// and store it in a rand. access temporary cache. 

jack_nframes_t ffjb_position = 0;
audiocache *ffjb_cache = NULL;
int ffjb_fadeedges = 0;  // fade in after padded frame - smooth edges - don't zero pad when buffering after resample
void *bufferthread(void *arg); 

/* ffmpeg buffer - read data into cache */
jack_nframes_t fillcl(audiocache *cache) {
	audiocache *cl;
	jack_nframes_t rv;
	
	cl= getcl(cache);
	cl->state&=~1; // TODO: make thread safe.
	rv=fillBuffer_ff(cl->d, CACHE_S);
	cl->end=ffjb_clock;
	cl->start=ffjb_clock;
	if (rv > ffjb_clock) { fprintf(stderr," invalid timestamp information !\n"); }
	else cl->start-= rv;
	cl->state|=1;
//	printf("read %i -> %i\n",cl->start,cl->end);
	return (rv);
}

jack_nframes_t fillBuffer_ffjb(jack_default_audio_sample_t **bufferptrs, jack_nframes_t frames )
{
	return(fillBuffer_cache(bufferptrs, frames, ffjb_position, ffjb_cache, &ffjb_fadeedges));
}

void sampleseek_ffjb (jack_nframes_t sample) {
	ffjb_position = sample;
}

void openfile_ffjb (char *filename) {
 	pthread_t pbufferthread;
	openfile_ff(filename);
	if (m_filetype == SFT_NONE) return;

	pthread_create(&pbufferthread,0,bufferthread,NULL);
}

// requires tweaking - until a proper semaphore/RT implementation is in place :)
void *bufferthread(void *arg){
	int completed=0;
	sampleseek_ff(0);
	ffjb_cache = newcl(NULL);

	//if (want_verbose) printf("started buffer-thread\n");

	while (1) {
		usleep(1);
		int direction = (want_speed<0)?-1:1;
		// TODO option: never clear cache   
		freecl(ffjb_position, ffjb_cache);
		// FIXME: sleep if completed until backwards seek.
		if (completed && ffjb_position > completed) { /* printf("Zzzzz.\n"); */ usleep(50000); continue; }
		if (testcls(ffjb_position, ffjb_cache) && 
			testcls(ffjb_position + direction*5*CACHE_S, ffjb_cache)) { usleep(50000); continue; }
		// FIXME: don'T seek beyond EOF 
//		if (ffjb_clock < ffjb_position )
		// TODO option: never seek: reset to pos(0) and decode sequentially.
		if ((ffjb_clock-11*CACHE_S > ffjb_position ) ||
		    (ffjb_clock+11*CACHE_S < ffjb_position )) { 
			printf("seek: from  %.0f to %i (req:%i)\n",ffjb_clock,((jack_nframes_t) ffjb_position/CACHE_S*CACHE_S) , ffjb_position);
			if(want_speed >=0) clearcache(ffjb_cache); // some files are just not reliably seekable :)
			sampleseek_ff((jack_nframes_t) (ffjb_position/CACHE_S*CACHE_S)-((want_speed <0)?(10*CACHE_S):0));
			completed=0;
		}

		if (!completed)  {
			if (fillcl(ffjb_cache) == 0 ) completed=1+ffjb_position;
		} else {  usleep(5000); }
	}
	return (NULL);
}

#endif /* HAVE_FFMPEG */


#ifdef HAVE_LQT
/////// QUICKTIME ///////
#include <lqt.h>
#include <quicktime.h>

quicktime_t *m_qt = NULL;

void openfile_qt (char *lqt_sucks_filename) {

	if ( !quicktime_check_sig( lqt_sucks_filename ) ) {
		fprintf(stderr, "This is not a Quicktime audio file\n" );
		return;
	}
	m_qt = quicktime_open( lqt_sucks_filename, 1, 0 );

	if ( !m_qt ) {
		fprintf(stderr, "Could not open Quicktime file\n" );
		return;
	}
	if ( quicktime_audio_tracks( m_qt ) == 0 ) {
		fprintf(stderr, "This Quicktime file does not have a audio track\n" );
		return;
	}
	if ( !quicktime_supported_audio( m_qt, 0 ) ) {
		fprintf(stderr, "quicktime Audio Codec is not supported\n" );
		return;
	}
	m_channels = quicktime_track_channels( m_qt, 0 );
	m_filetype = SFT_QUICKTIME;
//	fprintf(stderr, "long length? %li\n",quicktime_audio_length( m_qt, 0 ));

	m_length = quicktime_audio_length( m_qt, 0 );
	m_samplerate = quicktime_sample_rate ( m_qt, 0 );
	fprintf(stderr, "Sucessfully opened QT file len: %li trck: %i\n",quicktime_audio_length( m_qt, 0 ),
		quicktime_audio_tracks( m_qt ));
	
}
jack_nframes_t lqtpos = 0;

void sampleseek_qt (jack_nframes_t sample) {
//	fprintf(stderr, "seek: %i\n",sample);
//	fprintf(stderr, "DDD : file is %li - seek to %u \n",quicktime_audio_position( m_qt, 0 ),sample);
//	if (m_qt) quicktime_set_audio_position( m_qt, sample, 0 );
	lqtpos=sample;
}

jack_nframes_t fillBuffer_qt( jack_default_audio_sample_t ** bufferptrs, jack_nframes_t frames ) {
	int i;
//	int64_t diff = lqt_last_audio_position( m_qt, 0 );
//	if (m_qt) lqt_decode_audio_track( m_qt, NULL, bufferptrs, frames, 0 );
//	printf("PRE : file is %li - we want @%u len: %u\n",quicktime_audio_position( m_qt, 0 ),lqtpos,frames);
		quicktime_set_audio_position( m_qt, lqtpos, 0 );
//	printf("SEEK : file is %li - we want @%u len: %u\n",quicktime_audio_position( m_qt, 0 ),lqtpos,frames);
	for (i=0;m_qt && i<m_channels;i++) {
		quicktime_set_audio_position( m_qt, lqtpos, 0 );
		quicktime_decode_audio( m_qt, NULL, bufferptrs[i], frames, i );
	}
	return (frames);
//	return ((jack_nframes_t) (lqt_last_audio_position( m_qt, 0 ) - diff));
}

#endif /* HACE_LQT */


#ifdef HAVE_SNDFILE
/////// SNDFILE ///////
#include <sndfile.h>

SNDFILE * m_sndfile;
void openfile_sf (char *filename) {
	SF_INFO sfinfo;
	m_sndfile = sf_open( filename, SFM_READ, &sfinfo );	

	if ( SF_ERR_NO_ERROR != sf_error( m_sndfile ) ) {
		fprintf(stderr, "This is not a sndfile supported audio file format\n" );
		return;
	}
	if ( sfinfo.frames==0 ) {
		fprintf(stderr, "This is an empty audio file\n" );
		return;
	}
	m_channels = sfinfo.channels;
	m_samplerate = sfinfo.samplerate;
	m_length = (jack_nframes_t) sfinfo.frames;
	m_filetype = SFT_SNDFILE;
}

void sampleseek_sf (jack_nframes_t sample) {
	sf_seek( m_sndfile, sample, SEEK_SET );
}

float *interleaved = NULL;

void realloc_sf (jack_nframes_t buffersize) {
	if (interleaved) free(interleaved);
	interleaved=xcalloc(m_channels*buffersize, sizeof(float));
}

jack_nframes_t fillBuffer_sf( jack_default_audio_sample_t **bufferptrs, jack_nframes_t frames ) {
	int i,ii;
	jack_nframes_t nframes;
	nframes=sf_readf_float( m_sndfile, interleaved, frames);

	for (i=0;i<nframes && i< frames;i++) 
		for (ii=0;ii<m_channels;ii++) 
			bufferptrs[ii][i]= interleaved[m_channels*i+ii];
		
	return nframes;
}
#endif /* HAVE_SNDFILE */

