/* 
   jplay2 - JACK audio player for multimedia files

   Copyright (C) 2007 Robin Gareus

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 
   as published by the Free Software Foundation;

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include <stdio.h>
#include <sys/types.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h> // strtok
#include <strings.h> // strcasecmp
#include <signal.h>

#include "player.h"
#include "osc.h"

#define EXIT_FAILURE 1

extern int m_channels;
extern int m_samplerate;

static void usage (int status);

/* The name the program was run with, stripped of any leading path. */
char *program_name;

/* getopt_long return codes */
enum {DUMMY_CODE=129
};

/* Option flags and variables */

int want_scrub=0;	
int want_quiet=0;
int want_verbose=0;
int want_osc=0;
int want_transport=0;	
int jack_portc = -1; // -1: autoconnect ; > 0 number of ports to connect
char **jack_portnames = NULL;
float want_speed = 1.0;

// parameters
int p_scrublen=  3*1920; // will be rounded to jack bufsiz

static struct option const long_options[] =
{
  {"transport", no_argument, 0, 't'},
  {"scrub", no_argument, 0, 's'},
  {"speed", required_argument, 0, 'R'},
  {"ports", required_argument, 0, 'J'},
  {"jackports", required_argument, 0, 'J'},
  {"osc", required_argument, 0, 'o'},
  {"oscport", required_argument, 0, 'o'},
  {"scrublength", required_argument, 0, 'S'},
  {"quiet", no_argument, 0, 'q'},
  {"silent", no_argument, 0, 'q'},
  {"verbose", no_argument, 0, 'v'},
  {"help", no_argument, 0, 'h'},
  {"version", no_argument, 0, 'V'},
  {NULL, 0, NULL, 0}
};

static int decode_switches (int argc, char **argv);


void 
sig_byebye(int signal)
{
  if (want_osc) shutdown_osc();
  close_jack();
  exit(0);
}


int
main (int argc, char **argv)
{
  int i;

  program_name = argv[0];

  i = decode_switches (argc, argv);

  if (i != argc-1) usage(1);

  signal(SIGINT, sig_byebye);
  signal(SIGHUP, sig_byebye);
  signal(SIGTERM, sig_byebye);

  openfile(argv[i]);
  if (want_verbose) fprintf(stdout,"FILE-INFO: channels:%i srate:%i\n",m_channels,m_samplerate);	
#ifdef HAVE_LIBLO
  if (want_osc) initialize_osc();
#endif
  initialize_decoder(); // once jack is started m_channls MUST NOT change!
  dothejack();
  loopthejack();
  //loopthedebug();

#ifdef HAVE_LIBLO
  if (want_osc) shutdown_osc();
#endif
  exit (0);
}

static void printversion (void) {
  printf ( " compile time options: "
#ifdef HAVE_LIBLO
  "LIBLO "
#endif
#ifdef HAVE_FFMPEG
  "FFMPEG "
#endif
#ifdef HAVE_SNDFILE
  "SNDFILE "
#endif
#ifdef HAVE_LQT
  "QUICKTIME "
#endif
  "\n"
  );
}
/* Set all the option flags according to the switches specified.
   Return the index of the first non-option argument.  */

static int
decode_switches (int argc, char **argv)
{
  int c;
  char *jack_pn; 

  while ((c = getopt_long (argc, argv, 
			   "t"	/* transport */
			   "s"	/* scrub */
			   "o:"	/* osc */
			   "S:"	/* scrub */
			   "R:"	/* speed */
			   "J:"	/* jack ports */
			   "q"	/* quiet or silent */
			   "v"	/* verbose */
			   "h"	/* help */
			   "V",	/* version */
			   long_options, (int *) 0)) != EOF)
    {
      switch (c)
	{
	case 's':
	  want_scrub =1;
	  break;
	case 'S':
	  p_scrublen =atoi(optarg);
	  break;
	case 't':
	  want_transport =1;
	  break;
	case 'J':
	  jack_pn = strdup(optarg);
	  if (!strcasecmp(jack_pn,"none")) jack_portc=0;
	  else {
	    char *tmp;
	    tmp = strtok(jack_pn,",");
	    jack_portc=0;
	    while (tmp) {
	      jack_portnames=realloc(jack_portnames,(++jack_portc)*sizeof(char*));
	      jack_portnames[jack_portc-1]=strdup(tmp);
	      tmp = strtok(NULL,",");
	    }
	  }
	  free(jack_pn);
	  break;
	case 'R':
	  want_speed =atof(optarg);
	  //if (want_speed<=0.05) want_speed=0.05;
	  if (want_speed<-20.0) want_speed=-20.0;
	  if (want_speed>20.0) want_speed=20.0;
	  break;
	case 'q':		/* --quiet, --silent */
	  want_quiet = 1;
	  break;
	case 'o':
	  want_osc =atoi(optarg);
	  break;
	case 'v':		/* --verbose */
	  want_verbose = 1;
	  break;
	case 'V':
	  printf ("jplay2 %s\n", VERSION);
	  printversion();
	  exit (0);
	case 'h':
	  usage (0);
	default:
	  usage (EXIT_FAILURE);
	}
    }

  return optind;
}


static void
usage (int status)
{
  printf ("%s - \
JACK audio player for multimedia files\n", program_name);
  printf ("Usage: %s [OPTION]... [FILE]...\n", program_name);
  printf ("\
Options:\n\
  -t, --transport            play along jack transport\n\
  -s, --scrub                loop small region when transport is paused\n\
  -J <port>                  specify jack-ports to connect to or 'none'\n\
  -o <UDP-port>              start Open Sound Control server\n\
  -R <speed>                 0<n<20 <1:speedup >1: slow-down\n\
  -S <N>                     scrub N audioframes (rounded to jack bufsize)\n\
  -q, --quiet, --silent      inhibit usual output\n\
  -v, --verbose              print more information\n\
  -h, --help                 display this help and exit\n\
  -V, --version              output version information and exit\n\
");
  exit (status);
}
/* vi:set ts=8 sts=2 sw=2: */
